class VersionSchemaField:
    """String constants used as column names across snapshot table schemas."""

    SCHEMA_VERSION = "schema_version"
    TABLES = "tables"
    NAME = "name"
    PATH = "path"
    ROW_COUNT = "row_count"
    SRC_DATASET_ID = "src_dataset_id"
    PARENT_SRC_DATASET_ID = "parent_src_dataset_id"
    FULL_PATH = "full_path"
    DESCRIPTION = "description"
    SRC_VIDEO_ID = "src_video_id"
    SRC_VOLUME_ID = "src_volume_id"
    HASH = "hash"
    LINK = "link"
    FRAMES_COUNT = "frames_count"
    FRAME_WIDTH = "frame_width"
    FRAME_HEIGHT = "frame_height"
    FRAMES_TO_TIMECODES = "frames_to_timecodes"
    META = "meta"
    CUSTOM_DATA = "custom_data"
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"
    ANN_JSON = "ann_json"
    JSON = "json"
    ANNOTATION = "annotation"
    SRC_OBJECT_ID = "src_object_id"
    CLASS_NAME = "class_name"
    KEY = "key"
    TAGS_JSON = "tags_json"
    SRC_FIGURE_ID = "src_figure_id"
    FRAME_INDEX = "frame_index"
    GEOMETRY_TYPE = "geometry_type"
    GEOMETRY_JSON = "geometry_json"
