from supervisely.nn.artifacts.detectron2 import Detectron2
from supervisely.nn.artifacts.hrda import HRDA
from supervisely.nn.artifacts.mmclassification import MMClassification, MMPretrain
from supervisely.nn.artifacts.mmdetection import MMDetection, MMDetection3
from supervisely.nn.artifacts.mmsegmentation import MMSegmentation
from supervisely.nn.artifacts.ritm import RITM
from supervisely.nn.artifacts.rtdetr import RTDETR
from supervisely.nn.artifacts.unet import UNet
from supervisely.nn.artifacts.yolov5 import YOLOv5, YOLOv5v2
from supervisely.nn.artifacts.yolov8 import YOLOv8
